// dsk2hfe.cpp : Defines the entry point for the console application.

#include <iostream>
#include <cctype>
#include <algorithm>
#include <string>

using namespace std;

// Interface mode constants
#define IBMPC_DD_FLOPPYMODE				0x00
#define IBMPC_HD_FLOPPYMODE				0x01
#define ATARIST_DD_FLOPPYMODE			0x02
#define ATARIST_HD_FLOPPYMODE			0x03
#define AMIGA_DD_FLOPPYMODE				0x04
#define AMIGA_HD_FLOPPYMODE				0x05
#define CPC_DD_FLOPPYMODE				0x06
#define GENERIC_SHUGART_DD_FLOPPYMODE	0x07
#define IBMPC_ED_FLOPPYMODE				0x08
#define MSX2_DD_FLOPPYMODE				0x09
#define C64_DD_FLOPPYMODE				0x0A
#define EMU_SHUGART_FLOPPYMODE			0x0B
#define S950_DD_FLOPPYMODE				0x0C
#define S950_HD_FLOPPYMODE				0x0D
#define DISABLE_FLOPPYMODE				0xFE

// Encoding constants
#define ISOIBM_MFM_ENCODING				0x00
#define AMIGA_MFM_ENCODING				0x01
#define ISOIBM_FM_ENCODING				0x02
#define EMU_FM_ENCODING					0x03
#define UNKNOWN_ENCODING				0xFF

// DSK/AIM/HFE constants
#define dsk_buffer_size 860160	// DSK buffer size
#define dsk_track_len 5376		// DSK track lenght
#define aim_buffer_size 2068480	// buffer size for AIM data
#define track_len 12928			// AIM/MFM track lenght
#define sector_size 256			// Sector size in bytes
#define mfm_buffer_size 2068480	// MFM buffer size
#define hfe_track_len 26112		// HFE track lenght
#define hfe_block_size 512		// HFE block size
#define hfe_buffer_size 2088960	// HFE buffer size
#define hfe_file_size 2089984	// HFE entire file size
#define reserved_size 486		// count of reserved bytes

string fname, filename, file_input_ext, file_output_ext;
string filename_dsk, filename_aim, filename_hfe;

unsigned int a, b, i, dsk_offset, aim_offset, raw_offset, raw_track_len, dsk_size, aim_size, offset;
unsigned char volume, track, sector, byte, byte1, byte2, byte3, byte4, bit, bit_prev, bit_num, desync_bit, command;

unsigned char *dsk_buffer, *tmp_buffer, *raw_buffer, *mfm_buffer, *hxc_track_list, *hfe_file_buffer, *hfe_buffer, *aim_buffer, *aim_track;
FILE *f_input, *f_output;

char *input_filename_ext;
char *output_filename_ext;
char *arg_fname_input;
char *arg_fname_output;

const char welcome_msg[] = "\nDSK/AIM to HFE CLI Tool ";
const char     version[] = "v1.70";
const char       build[] = ".24.1204 ";
const char     edition[] = "'Agath' Edition by Jeeno Laretti";

struct hxcpicfileheader
{
	char           HEADERSIGNATURE[9];			// "HXCPICFE" signature include zero ended as 'Revision 0' byte
	// unsigned  char formatrevision = 0;		// Revision 0
	unsigned  char number_of_track;				// Number of track in the file
	unsigned  char number_of_side;				// Number of valid side (Not used by the emulator)
	unsigned  char track_encoding;				// Track Encoding mode
												// (Used for the write support - Please see the list above)
	unsigned short bitRate;						// Bitrate in Kbit/s. Ex : 250=250000bits/s
												// Max value : 500
	unsigned short floppyRPM;					// Rotation per minute (Not used by the emulator)
	unsigned  char floppyinterfacemode;			// Floppy interface mode. (Please see the list above.)
	unsigned  char dnu;							// Free
	unsigned short track_list_offset;			// Offset of the track list LUT in block of 512 bytes
												// (Ex: 1 = 512)
	unsigned  char write_allowed;				// The Floppy image is write protected ?
	unsigned  char single_step;					// 0xFF : Single Step  0x00 Double Step mode
	unsigned  char track0s0_altencoding;		// 0x00 : Use an alternate track_encoding for track 0 Side 0
	unsigned  char track0s0_encoding;			// alternate track_encoding for track 0 Side 0
	unsigned  char track0s1_altencoding;		// 0x00 : Use an alternate track_encoding for track 0 Side 1
	unsigned  char track0s1_encoding;			// alternate track_encoding for track 0 Side 1
	unsigned  char reserved[reserved_size];		// Reserved area
};

struct hxcpicfileheader hxcfileheader;
struct hxcpicfileheader *hxc_header = &hxcfileheader;

// ========================================================================================================================================================
void push(unsigned char data_byte, unsigned char cmd_byte, unsigned int count)
{
	for (count; count > 0; count--)
	{
		aim_track[aim_offset++] = data_byte;
		aim_track[aim_offset++] = cmd_byte;
	}
}
// ========================================================================================================================================================
void build_aim_from_dsk(void)
{
	// Expanding DSK to AIM
	aim_buffer = (unsigned char *)malloc(aim_buffer_size);

	// Fill AIM buffer with dummy sync data
	for (i = 0; i < aim_buffer_size;)
	{
		aim_buffer[i++] = 0xAA;
		aim_buffer[i++] = 0x00;
	}

	// Build AIM track from DSK data
	for (track = 0; track < 160; track++)
	{
		aim_track = (unsigned char *)malloc(track_len);
		aim_offset = 0;

		// GAP1: 144 bytes 0xAA
		push(0xAA, 0, 144);

		for (sector = 0; sector < 21; sector++)
		{
			dsk_offset = track * dsk_track_len + sector * sector_size;
			// DESYNC
			push(0x00, 1, 1);

			// INDEX mark
			push(0x95, 0, 1);
			push(0x6A, 0, 1);

			// Volume, Track, Sector
			push(0xFE, 0, 1);	// Volume, Default = 0xFE
			push(track, 0, 1);	// Track
			push(sector, 0, 1);	// Sector

			// End of INDEX mark
			push(0x5A, 0, 1);

			//GAP2: 3 bytes 0xAA
			push(0xAA, 0, 3);

			// DESYNC
			push(0x00, 1, 1);

			// DATA mark
			push(0x6A, 0, 1);
			push(0x95, 0, 1);

			// DATA
			unsigned short crc = 0;
			for (unsigned int count = 0; count < 256; count++)
			{
				byte = dsk_buffer[dsk_offset + count];
				if (crc > 0xFF)
				{
					crc++;
					crc &= 0xFF;
				}
				crc += byte;
				push(byte, 0, 1);
			}

			// CRC
			push(crc & 0xFF, 0, 1);

			// End of DATA
			push(0x5A, 0, 1);

			// Volume, Default = 0xFE
			// push(0xFE, 0, 1); // --- ???

			// GAP3: 28 bytes 0xAA
			push(0xAA, 0, 29);
		}
		// GAP4: 20 bytes 0xAA
		push(0xAA, 0, 19);
		
		// END of TRACK mark
		push(0xAA, 2, 1);
		// Total exactly 12928 bytes per track

		// Put track into AIM buffer
		for (i = 0; i < track_len; i++) aim_buffer[track * track_len + i] = aim_track[i];
	}
}
// ========================================================================================================================================================
void rebuild_aim_data(void)
{
	// Create TMP buffer
	tmp_buffer = (unsigned char *)malloc(aim_buffer_size);

	for (track = 0; track < 160; track++)
	{
		aim_offset = offset = track * track_len;
		i = 0;
		while (i < track_len)
		{
			byte1 = aim_buffer[aim_offset + 0];
			byte2 = aim_buffer[aim_offset + 1];
			if ((byte1 == 0x00 || byte1 == 0xA4) && byte2 == 1)
			{
				tmp_buffer[i + offset + 0] = 0xA4;
				tmp_buffer[i + offset + 1] = 0x01;
				tmp_buffer[i + offset + 2] = 0xFF;
				tmp_buffer[i + offset + 3] = 0x00;
				aim_offset += 2;
				i += 4;
			}
			else
			{
				tmp_buffer[i + offset + 0] = byte1;
				tmp_buffer[i + offset + 1] = byte2;
				aim_offset += 2;
				i += 2;
			}
		}
	}
	for (offset = 0; offset < aim_buffer_size; offset++) aim_buffer[offset] = tmp_buffer[offset];
}
// ========================================================================================================================================================
void build_mfm_stream(void)
{
	// Build MFM code
	const unsigned char mfm_encode[4] = { 64, 128, 0, 128 };

	mfm_buffer = (unsigned char *)malloc(mfm_buffer_size);

	for (track = 0; track < 160; track++)
	{
		bit_prev = 0;
		for (i = 0; i < track_len; i++)
		{
			a = track * track_len;
			byte = aim_buffer[a + i];
			command = aim_buffer[a + i + 1];

			if ((byte == 0x00 || byte == 0xA4) && command == 0x01)	// Desync
			{
				mfm_buffer[a + i++] = 0x22;
				mfm_buffer[a + i]   = 0x09;
			}
			else
			{
				bit_num = 0;
				mfm_buffer[a + i] = 0;
				while (bit_num++ < 8)
				{
					bit = (byte & 0x80) >> 7;
					byte = byte << 1;
					if (bit_num == 5)
					{
						i++;
						mfm_buffer[a + i] = 0;
					}
					else mfm_buffer[a + i] = mfm_buffer[a + i] >> 2;
					mfm_buffer[a + i] = mfm_buffer[a + i] | mfm_encode[bit_prev << 1 | bit];
					bit_prev = bit;
				}
			}
		}
	}
}
// ========================================================================================================================================================
int build_hfe_file(void)
{
	// Create HxC header
	std::string str = "HXCPICFE";
	strcpy_s(hxcfileheader.HEADERSIGNATURE, str.c_str());
	hxcfileheader.number_of_track = 80;
	hxcfileheader.number_of_side = 2;
	hxcfileheader.track_encoding = ISOIBM_MFM_ENCODING;
	hxcfileheader.bitRate = 250;
	hxcfileheader.floppyRPM = 0;
	hxcfileheader.dnu = 0;
	hxcfileheader.floppyinterfacemode = GENERIC_SHUGART_DD_FLOPPYMODE;
	hxcfileheader.track_list_offset = 1;
	hxcfileheader.write_allowed = 0;
	hxcfileheader.single_step = 0xFF;
	hxcfileheader.track0s0_altencoding = 0xFF;
	hxcfileheader.track0s0_encoding = 0xFF;
	hxcfileheader.track0s1_altencoding = 0xFF;
	hxcfileheader.track0s1_encoding = 0xFF;
	hxcfileheader.reserved[486] = 0;

	// Create HxC header
	std::fill_n(hxcfileheader.reserved, sizeof(hxcfileheader.reserved), 0xFF);

	// Rebuild AIM buffer
	rebuild_aim_data();
	
	// Build MFM data stream
	build_mfm_stream();

	//cout << "Writting ... MFM ... ";
	//fopen_s(&f_output, filename_mfm.c_str(), "wb");
	//fwrite(mfm_buffer, mfm_buffer_size, 1, f_output);
	//fclose(f_output);
	//cout << "Done\n";
	//return 0;

	// Create track list
	hxc_track_list = (unsigned char *)malloc(hfe_block_size);
	for (i = 0; i < hfe_block_size; i++) hxc_track_list[i] = 0xFF;
	offset = 2; // Start offset 2x512 bytes blocks

	a = 0;
	for (track = 0; track < 80; track++)
	{
		hxc_track_list[a++] = offset & 0xFF;
		hxc_track_list[a++] = (offset >> 8) & 0xFF;
		hxc_track_list[a++] = (track_len * 2) & 0xFF;
		hxc_track_list[a++] = ((track_len * 2) >> 8) & 0xFF;
		offset += hfe_track_len / hfe_block_size;
	}

	// Create HFE buffer
	hfe_buffer = (unsigned char *)malloc(hfe_buffer_size);

	// Create TMP buffer
	tmp_buffer = (unsigned char *)malloc(hfe_buffer_size);
	// fill with dummy sysnc data
	//for (i = 0; i < hfe_buffer_size; i++) tmp_buffer[a++] = 0x22;
	for (i = 0; i < hfe_buffer_size; i++) tmp_buffer[i] = 0x22;

	// Expand MFM to TMP bufer
	// MFM = 12928 bytes lenght
	// HFE = 13056 bytes lenght
	for (track = 0; track < 160; track++)
	{
		a = track * hfe_track_len / 2;
		b = track * track_len;
		for (i = 0; i < track_len; i++) tmp_buffer[a++] = mfm_buffer[b++];
	}

	// Assemble MFM data from TMP buffer into HFE buffer
	unsigned int n = 0;
	for (track = 0; track < 80; track++) // 2 sides with 12544 bytes per track
	{
		for (i = 0; i < hfe_track_len / hfe_block_size; i++) // number of blocks by 256 bytes
		{
			// Move two data block: first for Side 0, second for Side 1
			for (a = 0; a < 256; a++)
			{
				hfe_buffer[n + a] = tmp_buffer[(track * hfe_track_len) + (i * 256) + a];
				hfe_buffer[n + a + 256] = tmp_buffer[(track * hfe_track_len) + (i * 256) + a + hfe_track_len / 2];
			}
			n += hfe_block_size;
		}
	}

	cout << "Writting ... HFE ... ";
	fopen_s(&f_output, filename_hfe.c_str(), "wb");
	if (f_output == NULL)
	{
		cout << "Write error\n";
		return -1;
	}

	// Write HxC header
	fwrite(hxc_header, hfe_block_size, 1, f_output);

	// Write track list
	fwrite(hxc_track_list, hfe_block_size, 1, f_output);

	// Write HFE data buffer
	fwrite(hfe_buffer, hfe_buffer_size, 1, f_output);
	fclose(f_output);
	cout << "Done\n";

	return 0;
}

// ========================================================================================================================================================
void extract_mfm_data_from_hfe(void)
{
	// Create temporary buffer with size of HFE data size
	tmp_buffer = (unsigned char *)malloc(hfe_buffer_size);
	unsigned char blocks_count = hfe_track_len / hfe_block_size;

	// Extract MFM data from HFE body
	for (track = 0; track < 80; track++) // 160 Tracks /2 Sides
	{
		i = track * 4 + hfe_block_size;
		unsigned int block_address = (hfe_file_buffer[i] | hfe_file_buffer[i + 1] << 8) * hfe_block_size;
		for (unsigned char block_num = 0; block_num < blocks_count; block_num++)
		{
			for (a = 0; a < 256; a++)
				tmp_buffer[track * hfe_track_len + block_num * 256 + a] = hfe_file_buffer[block_address++];
			for (a = 0; a < 256; a++)
				tmp_buffer[track * hfe_track_len + block_num * 256 + a + hfe_track_len / 2] = hfe_file_buffer[block_address++];
		}
	}

	// Create MFM buffer
	mfm_buffer = (unsigned char *)malloc(mfm_buffer_size);

	// Compact TMP buffer to MFM buffer
	for (track = 0; track < 160; track++)
	{
		a = track * hfe_track_len / 2;
		for (i = 0; i < track_len; i++) mfm_buffer[track * track_len + i] = tmp_buffer[a++];
	}
}
// ========================================================================================================================================================
void decode_mfm_to_raw(void)
{
	// Decode MFM to useful data
	raw_buffer = (unsigned char *)malloc(mfm_buffer_size / 2);

	for (track = 0; track < 160; track++)
	{
		unsigned int raw_offset = track * track_len / 2;
		unsigned int mfm_offset = track * track_len;
		for (i = 0; i < track_len / 2; i++)
		{
			// Bits shifting
			byte = 0;
			b = (mfm_buffer[mfm_offset] & 0xFF) | (mfm_buffer[mfm_offset + 1] << 8);
			for (unsigned char x = 0; x < 16; x++)
			{
				if (x & 1) byte = (byte << 1) | (b & 1);
				b >>= 1;
			}
			raw_buffer[raw_offset++] = byte;
			mfm_offset += 2;
		}
	}
}
// ========================================================================================================================================================
void compct_aim_to_raw(void)
{
	// Create RAW buffer
	raw_buffer = (unsigned char *)malloc(aim_buffer_size / 2);

	aim_offset = 0;
	unsigned int raw_offset = 0;
	for (track = 0; track < 160; track++)
	{
		for (i = 0; i < track_len / 2; i++)
		{
			byte1 = aim_buffer[aim_offset++];
			byte2 = aim_buffer[aim_offset++];
			if ((byte1 == 0x00 || byte1 == 0xA4) && byte2 == 0x01) raw_buffer[raw_offset++] = 0xA4;
			else raw_buffer[raw_offset++] = byte1;
		}
	}
}

// ========================================================================================================================================================
// ========================================================================================================================================================
// ========================================================================================================================================================
// ========================================================================================================================================================
// ========================================================================================================================================================

int main(int argc, char *argv[])
{
	cout << welcome_msg << version << build << edition << "\n";
	cout << "2018-2024 (C) United Engineers Ltd.\n";

	if (argc < 2) return 0;
	else if (argc > 3)
	{
		cout << "Syntax error\n";
		return -1;
	}

	arg_fname_input  = argv[1];
	input_filename_ext = strrchr(arg_fname_input, '.');

	if (input_filename_ext != NULL)
	{
		// Looking for input file extension
		file_input_ext = input_filename_ext;
		transform(file_input_ext.begin(), file_input_ext.end(), file_input_ext.begin(), tolower);	// convert to lower case

		arg_fname_output = argv[2];
		output_filename_ext = strrchr(arg_fname_output, '.');

		// Looking for output file extension
		if (output_filename_ext != NULL)
		{
			if (file_input_ext != ".dsk" && file_input_ext != ".aim")
			{
				cout << "Incorrect input file name\n";
				return -1;
			}

			// Find extension of output file
			file_output_ext = output_filename_ext;
			transform(file_output_ext.begin(), file_output_ext.end(), file_output_ext.begin(), tolower);	// convert to lower case
			if (file_output_ext != ".hfe")
			{
				cout << "Incorrect output file name\n";
				return -1;
			}
		}
	}
	
	if (input_filename_ext == NULL || output_filename_ext == NULL)
	{
		cout << "Filename extension cannot be null\n";
		return -1;
	}

	// Open input file for reading
	fopen_s(&f_input, arg_fname_input, "rb");
	if (f_input == NULL)
	{
		cout << "File not found\n";
		return -1;
	}

	// Read input file
	cout << "Reading  ... ";
	fname = arg_fname_input;
	filename = fname.substr(0, fname.rfind("."));
	filename_dsk = filename + ".dsk";
	filename_aim = filename + ".aim";
	filename_hfe = filename + ".hfe";
	
// ========================================================================================================================================================

	// Build from '.dsk'
	if (file_input_ext == ".dsk")
	{
		// Determine file size
		fseek(f_input, 0, SEEK_END);
		dsk_size = ftell(f_input);
		fseek(f_input, 0, SEEK_SET);

		cout << "DSK ... ";
		if (dsk_size == dsk_buffer_size) cout << "840K\n";

		else if (dsk_size == dsk_buffer_size + 256)
		{
			cout << "840K with 256 byte header\n";
			fseek(f_input, 256, SEEK_SET);
			dsk_size -= 256;
		}
		else if (dsk_size == dsk_buffer_size + 128)
		{
			cout << "840K with 128 byte epilogue\n";
			dsk_size -= 128;
		}
		else if (dsk_size == dsk_buffer_size + 4)
		{
			cout << "840K with 4 byte epilogue\n";
			dsk_size -= 4;
		}
		else
		{
			fclose(f_input);
			cout << "Incorrect file size\n";	// File size is incorrect
			return -1;
		}

		// Read '.dsk' file
		// Create buffer and read DSK file
		dsk_buffer = (unsigned char *)malloc(dsk_size);
		fread(dsk_buffer, dsk_size, 1, f_input);
		fclose(f_input);

		// Expand DSK data to AIM
		build_aim_from_dsk();

		// Write '.aim' file
		if (file_output_ext == ".aim")
		{
			cout << "Writting ... AIM ... ";
			fopen_s(&f_output, filename_aim.c_str(), "wb");
			if (f_output == NULL)
			{
				cout << "Write error\n";
				return -1;
			}

			fwrite(aim_buffer, aim_buffer_size, 1, f_output);
			fclose(f_output);
			cout << "Done\n";
			return 0;
		}

		// Write '.hfe' file
		else
		{
			if (build_hfe_file() == -1)
			{
				cout << "Write error\n";
				return -1;
			}
		}
	}

// ========================================================================================================================================================

	// Build from '.aim'
	else if (file_input_ext == ".aim")
	{
		// Determine file size
		fseek(f_input, 0, SEEK_END);
		aim_size = ftell(f_input);
		fseek(f_input, 0, SEEK_SET);

		cout << "IAM ... ";
		if (aim_size == aim_buffer_size) cout << "840K\n";
		else
		{
			fclose(f_input);
			cout << "Incorrect file size\n";
			return -1;
		}

		// Create buffer and read AIM file into
		aim_buffer = (unsigned char *)malloc(aim_buffer_size);
		fread(aim_buffer, aim_size, 1, f_input);
		fclose(f_input);

		// AIM - Create '.hfe' file
		if (file_output_ext == ".hfe")
		{
			if (build_hfe_file() == -1)
			{
				cout << "Write error\n";
				return -1;
			}
		}
	}

// ========================================================================================================================================================

	// If parameters are not set correctly
	else
	{
		cout << "Invalid input file type\n";
		return -1;
	}

// ========================================================================================================================================================


	// Global return
	return 0;
}
